# CORNish PASDy -- COsmic-Ray Neutron flavored Processing and Analysis of Sensor Data in pYthon

This is a project developed by Martin Schrön,  Erik Nixdorf, Jannis Jakobi, Mandy Kasner, Rafael Rosolem,
in order to make Cosmic Ray Neutron Sensing Data Analysis accessible for everyone!
Please ask us about coauthorship if you use our code for a publication.

*Note:* This version of PASDy is specific for Cosmic-Ray Neutron Sensing (CRNS). A more general version of PASDy is to be developed in the far future.

## Contribution

- Did you find a bug or are you missing a feature? Please report your experience in the [project issues](https://git.ufz.de/CRNS/cornish_pasdy/issues).
- Have you resolved errors or added new features? Please push your contributions back to GitLab such that the whole community can benefit from your work. Contact [Martin](mailto:martin.schroen@ufz.de) to receive write permission.

## Usage

The CORNish PASDy package aims to allow various ways of usage depending on the level of detail and interaction required by individual users. It can be used in the following ways:

- **Mouse click** on the file `instantPASDy.py` processes the sensor data instantly. It requires a file `config.cfg` to be present locally. Use this option if you have only one type of sensor for which the exact same analysis needs to be performed over and over again.
- In the **console**, type
    
        python instantPASDy.py path/to/your/configfile.cfg
    
    Use this option if you don't like GUIs or you apply batch scripting. Add the parameter `--nonstop` to avoid the final input prompt at the end of the script.
- **Execute** `instantPASDy-GUI.py` where the config file can be selected and ammended directly through the interface.
- Work on your own scripts (e.g., using Jupyter notebooks) and **import** the required snippets from the PASDy package.  
    *(This option is currently under development)*
   

## Install

### Python 3.*

Has been tested on:

- Windows
    - [Python.org](https://www.python.org/downloads/) distribution 
        - version `3.7.*`

Does not seem to work on: 

- with Anaconda's python 3.6 on Windows (installation issues with package `osmnx`)


### Use Virtual Environments

The concept of virtual environments protects your other python projects from interfering package versions and dependency clashes.
It is also recommended for Cornish Pasdy. 

To create a new environment after cloning Cornish Pasdy, go to the repository folder and execute:

    python -m venv env
    
Now you can install packages and mess around without disturbing your other python projects. To start working, activate the session with:

    source env/Scripts/activate
    
If the execution fails on Windows, it's probably a line-ending error. Edit the file `env/Scripts/activate` and change line endings to `LF` (Unix).

To deactivate the session, just type `deactivate`.
    
### Packages

You have two options here:

- Find missing packages that you need to install or update. For this, once install the `pipreq` package and execute:
    
        pipreqs ./
    
    The list of required packages is saved to `requirements.txt`. 
    
- Use the pre-defined package list deployed by the project in `env/requirements.txt`.
    It probably doesn't contain the latest package versions, but it makes sure that the package versions you install are compatible with Cornish Pasdy.

Depending on the option you choose, start the installation with...

    pip install -r requirements.txt
    # or
    pip install -r env/requirements.txt
    
**Got errors?** If the installation of some packages fails,
try to download and install the corresponding wheel files directly from
<https://www.lfd.uci.edu/~gohlke/pythonlibs/> and install each of them with:

    pip install name_of_downloaded_file.whl
    
Experience has shown that there are lots of issues and dependencies with GDAL.
If so, first download and install the wheel files for `GDAL`, `Fiona`, `Cartopy`, `Rasterio`, `Rtree`
before continuing with the other packages the normal way.

Please remove the file `requirements.txt` (but not the one in `env/`) after you are done, as it is user-specific and should not be part of the repo.

### Jupyter Lab

Jupyter notebooks are a great way to play with python, here is a short instruction of installation:

    pip install jupyter
    pip install jupyterlab
    jupyter nbextension enable --py --sys-prefix widgetsnbextension
    jupyter labextension install @jupyterlab/git
    pip install --upgrade jupyterlab-git
    jupyter serverextension enable --py jupyterlab_git
      
Apply beautiful styles:

1. copy the file `docs/custom.css` into your home folder to `~/.jupyter/custom/`
2. copy the file `docs/pasdy.mplstyle` into your home folder to `~/.matplotlib/stylelib/`
3. Install the `Handlee` font family: <https://fonts.google.com/specimen/Handlee?selection.family=Handlee>

Run Jupyter Lab:

    jupyter lab


## Examples

Example station data:

    ./instantPASDy.py test/input/station.cfg

Example rover data:

    ./instantPASDy.py test/input/rover.cfg

Note that test comparison with reference data currently is not supported.

